import numpy
import pandas  # type: ignore
from lenu.data.elf_codes import ELFAbbreviations
from sklearn.base import TransformerMixin  # type: ignore


class ELFAbbreviationTransformer(TransformerMixin):
    def __init__(
        self,
        elf_abbreviations: ELFAbbreviations,
        use_endswith=True,
        use_lowercasing=True,
    ):
        self.elf_abbreviations = elf_abbreviations
        self.use_endswith = use_endswith
        self.use_lowercasing = use_lowercasing

    def _get_jurisdiction_from_input(self, X):
        return X["Jurisdiction"].iloc[0]

    def fit(self, X, y=None):
        # does nothing, but needs to be implemented
        return self

    def transform(self, X, y=None):
        jurisdiction = self._get_jurisdiction_from_input(X)

        abbreviations = self.elf_abbreviations.abbreviations_for_jurisdiction(
            jurisdiction
        )

        predictions = []
        for name in X["Entity.LegalName"]:
            pred = []  # list can end up empty
            for abbr in abbreviations:
                if self.elf_abbreviations.matches(
                    name, abbr, self.use_lowercasing, self.use_endswith
                ):
                    pred.append(1)
                else:
                    pred.append(0)

            predictions.append(pred)  # list can be empty
        return pandas.DataFrame(
            numpy.array(predictions),
            columns=[f"abbr({abbr})" for abbr in abbreviations],
        )

    def get_params(self, **kwargs):
        return {
            "elf_abbreviations": self.elf_abbreviations,
            "use_endswith": self.use_endswith,
            "use_lowercasing": self.use_lowercasing,
        }



# coding=utf-8

import unicodedata
import numpy


def _replace_diacritics(s):
    # lambda x : x.replace(u'ø', u'o').replace(u'æ', u'ae'),
    # what about ü,ö,ä,...
    return unicodedata.normalize("NFKD", s).encode("ASCII", "ignore").decode("utf-8")


def _replace_multi_spaces(s):
    while "  " in s:
        s = s.replace("  ", " ")
    return s


def _replace_double_quotation_marks(s):
    return s.replace('"', " ")


def _replace_trailing_non_alphanumeric_chars(s):
    while s and not s[-1].isalnum():
        s = s[0:-1]
    return s


def _correct_commas_and_periods(s):
    s = s.replace(" ,", ",")
    s = s.replace(" .", ".")
    s = s.replace(", ", ",")
    # s = s.replace(u'. ', u'.')
    return s


# def _replace_spelling_variations(s):
#    return s

purge_rules = [
    lambda x: x.replace(" l'", " l "),  # french example: 'L'Habitat'
    lambda x: x.replace("-", " "),
    lambda x: x.replace("(", " "),
    lambda x: x.replace(")", " "),
    # lambda x : x[4:] if x.startswith(u'the ') else x,
    lambda x: x.replace(" & ", " and "),
    lambda x: x.replace(" + ", " and "),
    lambda x: x.replace(";", " "),
    lambda x: x.replace("/", " "),
    lambda x: x.replace(",", " "),
]


def purge(s):
    for rule in purge_rules:
        s = rule(s)
    return s


steps = [
    lambda x: x.lower(),
    _replace_diacritics,
    _replace_multi_spaces,
    _replace_double_quotation_marks,
    _replace_trailing_non_alphanumeric_chars,
    _correct_commas_and_periods,
    purge,
    _replace_multi_spaces,
]


def harmonize(s):
    """
    Harmonization for company name matching inspired by
    http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.95.6455&rep=rep1&type=pdf
    """
    for step in steps:
        try:
            s = step(s)
        except Exception as e:
            raise Exception("Step failed: %s - %s" % (str(step), s), e)
    return s


_synonyms = {
    "&": "and",
    "ag": "aktiengesellschaft",
    "co": "company",
    "co.": "company",
    "corp": "corporation",
    "corp.": "corporation",
    "inc": "incorporated",
    "inc.": "incorporated",
    # u'inco': u'incorporated',
    "int": "international",
    "int.": "internatinal",
    "intl.": "international",
    "ltd": "limited",
    "ltd.": "limited",
    "pvt": "private",
}


def tokenize(s):

    # "Adding special case tokenization rules"
    # from https://spacy.io/usage/linguistic-features
    s = harmonize(s).strip()
    tokens = s.split(" ")

    # Spacy: use lemmatizer to put legal form surface forms into
    # lemmas (what I call synonyms here)
    # I imagine something like this: "Ltd." will be lemmatized as "limited".
    # https://spacy.io/usage/linguistic-features
    return numpy.unique([_synonyms.get(token, token) for token in tokens])



from pathlib import Path
import os
import urllib
from typing import Optional
import shutil

from lenu import data
from lenu.data.elf_codes import load_elf_code_list, ELFAbbreviations, ELF_CODE_FILE_NAME
from lenu.data.lei import (
    load_lei_cdf_data,
    COL_LEGALNAME,
    COL_JURISDICTION,
    COL_ELF,
    get_legal_jurisdiction,
)
from lenu.data.goldencopyfiles import (
    GoldenCopyFilePublications,
    GoldenCopyFilePublication,
)

from logging import getLogger

try:
    from importlib import resources
except ImportError:
    # backport for python < 3.9
    import importlib_resources as resources


logger = getLogger(__name__)


class DataRepoNotReady(Exception):
    pass


class DataRepo:
    def __init__(self, data_dir: Path):
        self.data_dir = data_dir

    def latest_lei_file(self) -> Optional[Path]:
        lei_files = list(
            sorted(self.data_dir.glob("*-gleif-goldencopy-lei2-golden-copy.csv.zip"))
        )
        latest_lei_file = lei_files[-1] if lei_files else None
        return latest_lei_file

    def elf_code_list_file(self) -> Optional[Path]:
        elf_file = Path(self.data_dir).joinpath(ELF_CODE_FILE_NAME)
        return elf_file if elf_file.exists() else None

    def ready(self) -> bool:
        return bool(self.latest_lei_file()) and bool(self.elf_code_list_file())

    def load_lei_cdf_data(self, jurisdiction):
        if not self.ready():
            raise DataRepoNotReady()

        logger.info(f"Loading LEI data into memory ({self.latest_lei_file()})")
        lei_data = load_lei_cdf_data(
            url=self.latest_lei_file(),
            usecols=[
                "LEI",
                COL_LEGALNAME,
                COL_JURISDICTION,
                COL_ELF,
                "Entity.LegalAddress.Region",
            ],
        ).assign(Jurisdiction=lambda d: d.apply(get_legal_jurisdiction, axis=1))
        jurisdiction_data = lei_data[lei_data["Jurisdiction"] == jurisdiction]
        return jurisdiction_data

    def load_elf_code_list(self):
        if not self.ready():
            raise DataRepoNotReady()

        logger.info(f"Loading ELF Code list ({self.elf_code_list_file()})")
        return load_elf_code_list(self.elf_code_list_file())

    def load_elf_abbreviations(self) -> ELFAbbreviations:
        elf_code_list = self.load_elf_code_list()
        elf_abbreviations = ELFAbbreviations.from_elf_code_list(elf_code_list)
        return elf_abbreviations

    def download_latest(self) -> None:
        publications = GoldenCopyFilePublications()
        publication: GoldenCopyFilePublication = publications.fetch_latest()

        lei_data_url = publication.lei2.full_file.csv.url
        filename = os.path.basename(lei_data_url)

        logger.info(f"Downloading {filename} to {self.data_dir}")
        urllib.request.urlretrieve(lei_data_url, self.data_dir.joinpath(filename))

        logger.info(f"Provide ELF Code list to {self.data_dir}")
        elf_target = self.data_dir.joinpath(ELF_CODE_FILE_NAME)
        with resources.path(data, ELF_CODE_FILE_NAME) as elf_resource:
            shutil.copy(elf_resource, elf_target)

    @staticmethod
    def from_data_dir(data_dir: Path) -> "DataRepo":
        if not data_dir.exists() or not data_dir.is_dir():
            raise ValueError(
                f"Given data_dir {str(data_dir)} does not exist or is not a directory."
            )
        return DataRepo(data_dir)



import pandas  # type: ignore
import numpy

from lenu.data.elf_codes import ELFAbbreviations
from lenu.ml.features import ELFAbbreviationTransformer


class TestELFAbbreviationsTransformer:
    def test_elf_abbreviation_transformer(self):
        elf_abbr = ELFAbbreviations(
            pandas.DataFrame(
                [
                    {"Jurisdiction": "DE", "ELF Code": "2HBR", "Abbreviation": "GmbH"},
                    {"Jurisdiction": "DE", "ELF Code": "40DB", "Abbreviation": "OHG"},
                    {
                        "Jurisdiction": "DE",
                        "ELF Code": "40DB",
                        "Abbreviation": "OHG mbH",
                    },
                ]
            )
        )

        elf_abbr_transformer = ELFAbbreviationTransformer(elf_abbr)

        res = elf_abbr_transformer.transform(
            pandas.DataFrame(
                {
                    "Entity.LegalName": ["Hallo GmbH", "Hello OHG mbH"],
                    "Jurisdiction": ["DE", "DE"],
                }
            )
        )

        assert (res == numpy.array([[1, 0, 0], [0, 0, 1]])).all().all()

        assert list(res.columns) == ["abbr(GmbH)", "abbr(OHG)", "abbr(OHG mbH)"]



import pandas  # type: ignore

# some columns names as constants for quick reuse
COL_LEGALNAME = "Entity.LegalName"
COL_LEGALNAME_LANG = "Entity.LegalName.xmllang"
COL_JURISDICTION = "Entity.LegalJurisdiction"
COL_ELF = "Entity.LegalForm.EntityLegalFormCode"


def load_lei_cdf_data(url, usecols=None):
    return pandas.read_csv(
        url,
        compression="zip",
        low_memory=False,
        dtype=str,
        # the following will prevent pandas from converting strings like 'NA' to NaN.
        na_values=[""],
        keep_default_na=False,
        usecols=usecols,
    )


def get_legal_jurisdiction(lei):
    if lei["Entity.LegalJurisdiction"] == "US":
        # this means we have a ISO-3166-2 code here
        if pandas.notnull(lei["Entity.LegalAddress.Region"]):
            return lei["Entity.LegalAddress.Region"]
        else:
            return lei["Entity.LegalJurisdiction"]
    else:
        return lei["Entity.LegalJurisdiction"]



import json
import logging
from datetime import datetime
from typing import Optional, List

import requests
from pydantic import BaseModel, Field, ValidationError

GLEIF_GOLDEN_COPY_URL = (
    "https://leidata-preview.gleif.org/api/v2/golden-copies/publishes"
)

logger = logging.getLogger(__name__)


class FileRef(BaseModel):
    type: str
    format: str
    record_count: int
    size: int
    size_human_readable: str
    delta_type: str
    url: str


class FileInVariousFormats(BaseModel):
    csv: FileRef
    json_: FileRef = Field(alias="json")
    xml: FileRef


class DeltaFiles(BaseModel):
    IntraDay: Optional[FileInVariousFormats]
    LastDay: Optional[FileInVariousFormats]
    LastWeek: Optional[FileInVariousFormats]
    LastMonth: Optional[FileInVariousFormats]


class GoldenCopyFileBundle(BaseModel):
    type: str
    publish_date: datetime
    full_file: FileInVariousFormats
    delta_files: DeltaFiles  # can be [] as well


class GoldenCopyFilePublication(BaseModel):
    publish_date: datetime
    lei2: GoldenCopyFileBundle
    rr: GoldenCopyFileBundle
    repex: GoldenCopyFileBundle


class GoldenCopyFilePublications:
    """
    Allows for fetching information about GLEIF Golden Copy files.
    """

    def __init__(self, url: str = GLEIF_GOLDEN_COPY_URL, page_size=10):
        self._url = url
        self._page_size = page_size

    def _fetch_page(self, page: int) -> List[GoldenCopyFilePublication]:
        logger.debug('Fetch list of golden copy files from "{0}"'.format(self._url))
        response = requests.get(
            self._url, params={"page": page, "page_size": self._page_size}
        )
        publications = response.json()["data"]

        result = []
        for p in publications:
            try:
                result.append(GoldenCopyFilePublication.parse_obj(p))
            except ValidationError:
                p_str = json.dumps(p, indent=2)
                logger.exception(f"could not parse \n {p_str}")
                raise
        return result

    def fetch_latest(self) -> GoldenCopyFilePublication:
        page_publications = self._fetch_page(page=1)
        return list(
            sorted(page_publications, key=lambda p: p.publish_date, reverse=True)
        )[0]



from pathlib import Path
import sys
from logging import getLogger
from typing import Optional

import typer
from typer import Typer, echo

from lenu.data import DataRepo

from lenu.ml.pipelines import ModelRepo
from lenu.util import typer_log_config

logger = getLogger(__name__)

DEFAULT_DATA_DIR = "./data"
DEFAULT_MODEL_DIR = "./models"


app = Typer(callback=typer_log_config)


@app.command()
def download(
    data_dir: Path = typer.Option(
        DEFAULT_DATA_DIR, exists=True, dir_okay=True, resolve_path=True
    )
):
    """
    Download latest Golden Copy LEI file(s) and ELF Code list in csv format.
    :param data_dir: where to store the files
    """

    data_repo = DataRepo.from_data_dir(data_dir)
    data_repo.download_latest()


@app.command()
def train(
    jurisdiction: str,
    data_dir: Path = typer.Option(
        DEFAULT_DATA_DIR, exists=True, dir_okay=True, resolve_path=True
    ),
    models_dir: Path = typer.Option(
        DEFAULT_MODEL_DIR, exists=True, dir_okay=True, resolve_path=True
    ),
):
    data_repo = DataRepo.from_data_dir(data_dir)
    model_repo = ModelRepo.from_models_dir(models_dir)

    if data_repo.ready():
        model_repo.train_pipeline(jurisdiction, data_repo)
    else:
        logger.error("LEI data is not ready yet, Please use `lenu download`")


@app.command()
def elf(
    jurisdiction: str,
    legal_name: str,
    data_dir: Path = typer.Option(
        DEFAULT_DATA_DIR, exists=True, dir_okay=True, resolve_path=True
    ),
    models_dir: Path = typer.Option(
        DEFAULT_MODEL_DIR, exists=True, dir_okay=True, resolve_path=True
    ),
):
    data_repo = DataRepo.from_data_dir(data_dir)

    if not data_repo.ready():
        logger.error("LEI data is not ready yet, Please use `lenu download`")
        sys.exit(1)

    model_repo = ModelRepo.from_models_dir(models_dir)

    try:
        elf_model = model_repo.get_model(jurisdiction)
    except ValueError as ve:
        logger.error(ve.args[0])
        sys.exit(1)

    elf_probabilities = elf_model.detect(legal_name, top=3)

    # map things back to ELF Code and present
    elf_codes = data_repo.load_elf_code_list().groupby("ELF Code").first()
    res = (
        elf_codes.loc[elf_probabilities.index]
        .assign(Score=elf_probabilities)
        .reset_index()
        .rename(columns={"index": "ELF Code"})[
            ["ELF Code", "Entity Legal Form name Local name", "Score"]
        ]
    )

    echo(res)


@app.command()
def abbreviations(
    jurisdiction: str,
    abbr: Optional[str] = None,
    data_dir: Path = typer.Option(
        DEFAULT_DATA_DIR, exists=True, dir_okay=True, resolve_path=True
    ),
):
    data_repo = DataRepo.from_data_dir(data_dir)

    if not data_repo.ready():
        logger.error("LEI data is not ready yet, Please use `lenu download`")
        sys.exit(1)

    elf_abbr = data_repo.load_elf_abbreviations()
    if abbr:
        echo(f'ELF Codes for "{abbr}" in Jurisdiction {jurisdiction}:')
        echo(elf_abbr.elf_codes_for_abbreviation(jurisdiction, abbr))
    else:
        echo(f"Abbreviations in Jurisdiction {jurisdiction}:")
        echo(elf_abbr.abbreviations_for_jurisdiction(jurisdiction))


if __name__ == "__main__":
    app()



import logging
from logging import Handler, LogRecord

from click import echo


class TyperEchoHandler(Handler):
    def emit(self, record: LogRecord) -> None:
        if record.levelno >= self.level:
            msg = (
                self.formatter.format(record) if self.formatter else record.getMessage()
            )
            echo(msg, err=record.levelno >= logging.WARNING)


def typer_log_config(verbose: bool = False, loglevel: str = "INFO"):
    if verbose:
        logger = logging.getLogger()
        logger.setLevel(loglevel.upper())

        formatter = logging.Formatter(
            "%(asctime)s | %(levelname)s | %(message)s", "%m-%d-%Y %H:%M:%S"
        )

        echo_handler = TyperEchoHandler()
        echo_handler.setFormatter(formatter)
        logger.addHandler(echo_handler)



import pandas  # type: ignore
from lenu.data.lei import COL_ELF
from sklearn.model_selection import (  # type: ignore
    cross_validate,
    StratifiedShuffleSplit,
)


def evaluate_jurisdiction(jurisdiction_data, pipeline):

    X = jurisdiction_data
    y = jurisdiction_data[COL_ELF]

    crossvalidation = StratifiedShuffleSplit(n_splits=10, test_size=0.3, random_state=0)
    # try:
    results = cross_validate(
        pipeline,
        X,
        y,
        cv=crossvalidation,
        scoring=[
            "accuracy",
            "balanced_accuracy",
            "f1_micro",
            "f1_macro",
            "f1_weighted",
        ],
        return_train_score=True,
        return_estimator=True,
    )
    return pandas.Series(results)



import pandas  # type: ignore
from lenu.data.elf_codes import ELFAbbreviations
from sklearn.base import (  # type: ignore
    BaseEstimator,
    ClassifierMixin,
)
from sklearn.utils.validation import check_is_fitted  # type: ignore


class ELFAbbreviationClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        elf_abbreviations: ELFAbbreviations,
        use_endswith=True,
        use_lowercasing=True,
    ):
        self.elf_abbreviations = elf_abbreviations
        self.use_endswith = use_endswith
        self.use_lowercasing = use_lowercasing

    def _get_jurisdiction_from_input(self, X):
        return X["Jurisdiction"].iloc[0]

    def fit(self, _, y):
        self.frequencies_ = pandas.value_counts(y)
        self.most_frequent_ = self.frequencies_.idxmax()
        return self

    def _best_elf_code_for_abbr(self, jurisdiction, name, abbr):
        if self.elf_abbreviations.matches(
            name, abbr, self.use_lowercasing, self.use_endswith
        ):
            elf_codes = self.elf_abbreviations.elf_codes_for_abbreviation(
                jurisdiction, abbr
            )
            pred = max(
                elf_codes, key=lambda elf_code: self.frequencies_.get(elf_code, 0)
            )
            return pred
        else:
            return None

    def predict(self, data):
        check_is_fitted(self, ["frequencies_", "most_frequent_"])

        jurisdiction = self._get_jurisdiction_from_input(data)

        abbreviations = self.elf_abbreviations.abbreviations_for_jurisdiction(
            jurisdiction
        )

        predictions = []
        for name in data["Entity.LegalName"]:
            potential_elf_codes = [
                self._best_elf_code_for_abbr(jurisdiction, name, abbr)
                for abbr in abbreviations
            ]

            pred = max(
                [elf_code for elf_code in potential_elf_codes if elf_code is not None],
                key=lambda elf_code: self.frequencies_.get(elf_code, 0),
                default=self.most_frequent_,
            )

            predictions.append(pred)
        return predictions



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lenu', 'lenu.data', 'lenu.ml', 'lenu.ml.test']

package_data = \
{'': ['*']}

install_requires = \
['importlib-resources>=5.7.1,<6.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'typer[all]>=0.4.1,<0.5.0',
 'types-requests>=2.27.16,<3.0.0']

entry_points = \
{'console_scripts': ['lenu = lenu.console:app']}

setup_kwargs = {
    'name': 'lenu',
    'version': '0.1.1',
    'description': 'Legal Entity Name Understanding',
    'long_description': '\n<h1 align="center">\nlenu - Legal Entity Name Understanding \n</h1>\n\n---------------\n\n<h1 align="center">\n<a href="https://gleif.org">\n<!-- Logo for github (dark background)\n<img src="http://sdglabs.ai/wp-content/uploads/2022/07/GLIEF-Logo-Colour-Reversed.png" width="220">\n-->\n<!-- Logo for pypi (white background) - uncomment when publishing to pypi -->\n<img src="http://sdglabs.ai/wp-content/uploads/2022/07/gleif-logo-new.png" width="220">\n</a>\n</h1><br>\n<h3 align="center">in collaboration with</h3> \n<h1 align="center">\n<a href="https://sociovestix.com">\n<img src="https://sociovestix.com/img/svl_logo_centered.svg" width="700px">\n</a>\n</h1><br>\n\n---------------\n\n[![License](https://img.shields.io/github/license/Sociovestix/lenu.svg)](https://github.com/Sociovestix/lenu/blob/main/LICENSE)\n![](https://img.shields.io/badge/python-3.8%20%7C%203.9-blue)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n\n**lenu** is a python library that helps to understand and work with Legal Entity Names\nin the context of the [Legal Entity Identifier](https://www.gleif.org/en/about-lei/introducing-the-legal-entity-identifier-lei) (LEI) Standard (ISO 17441)\nas well as the [Entity Legal Form (ELF) Code List](https://www.gleif.org/en/about-lei/code-lists/iso-20275-entity-legal-forms-code-list) Standard (ISO 20275).  \n\nThe library utilizes Machine Learning with scikit-learn and is distributed under Creative Commons Zero 1.0 Universal license.   \n\nThe project was started in November 2021 as a collaboration of the [Global Legal Entity Identifier Foundation](https://gleif.org) (GLEIF) and\n[Sociovestix Labs](https://sociovestix.com) with the goal to explore how Machine Learning can support in detecting the legal form (ELF Code) from a legal name. \n\nIt provides:\n- an interface to download [LEI](https://www.gleif.org/en/lei-data/gleif-golden-copy/download-the-golden-copy#/) and [ELF Code](https://www.gleif.org/en/about-lei/code-lists/iso-20275-entity-legal-forms-code-list) data from GLEIF\'s public website\n- an interface to train and make use of Machine Learning models to classify ELF Codes from given Legal Names\n\n---\n\n## Dependencies\n**lenu** requires\n- python (>=3.8, <3.10)\n- [scikit-learn](https://scikit-learn.org/) - Provides the Machine Learning functionality \n- [pandas](https://pandas.pydata.org/) - For reading and handling data\n- [Typer](https://typer.tiangolo.com/) - Adds the command line interface\n- [requests](https://docs.python-requests.org/en/latest/) and [pydantic](https://pydantic-docs.helpmanual.io/) - For downloading LEI data from GLEIF\'s website\n\n## Installation\n\nvia PyPI:\n```shell\npip install lenu\n```\n\nFrom github:\n```shell\npip install https://github.com/Sociovestix/lenu\n```\n\nEditable install from locally cloned repository\n```shell\ngit clone https://github.com/Sociovestix/lenu\npip install -e lenu\n```\n\n## Usage\n\nCreate folders for LEI and ELF Code data and to store your models\n\n```shell\nmkdir data\nmkdir models\n```\n\nDownload LEI data and ELF Code data into your `data` folder\n```shell\nlenu download\n```\n\nTrain a (default) ELF Code Classification model. An ELF Classification model is always Jurisdiction specific and \nwill be trained from Legal Names from this Jurisdiction.\n\nExamples: \n```shell\nlenu train DE       # Germany\nlenu train US-DE    # United States - Delaware\nlenu train IT       # Italy\n```\n\nIdentify ELF Code by using a model. The tool will return the best scoring ELF Codes. \n```shell\nlenu elf DE "Hans Müller KG"\n#   ELF Code                  Entity Legal Form name Local name     Score\n# 0     8Z6G                              Kommanditgesellschaft  0.979568\n# 1     V2YH                       Stiftung des privaten Rechts  0.001141\n# 2     OL20  Einzelunternehmen, eingetragener Kaufmann, ein...  0.000714\n```\n\n\n## Support and Contributing\nFeel free to reach out to either [Sociovestix Labs](https://sociovestix.com/contact) or [GLEIF](https://www.gleif.org/contact/contact-information)\nif you need support in using this library, in utilizing LEI data in general, or in case you would like to contribute to this library in any form.\n',
    'author': 'aarimond',
    'author_email': 'alexander.arimond@sociovestix.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from functools import lru_cache

import numpy
import pandas  # type: ignore

ELF_CODE_FILE_NAME = "2021-10-21-elf-code-list-v1.4.1.csv"


def load_elf_code_list(url):
    return pandas.read_csv(
        url,
        low_memory=False,
        dtype=str,
        # the following will prevent pandas from converting strings like 'NA' to NaN.
        na_values=[""],
        keep_default_na=False,
    )


def get_jurisdiction(elf_code_list):
    # We picked region code if we have one, otherwise Country Code
    return elf_code_list.apply(
        lambda x: x["Country sub-division code (ISO 3166-2)"]
        if pandas.notnull(x["Country sub-division code (ISO 3166-2)"])
        else x["Country Code (ISO 3166-1)"],
        axis=1,
    )


def expand_us_states(elf_code_list):
    # In US we have ELF Codes on federal level and ELF Codes on State level
    # We expand the list and for each state we add the federal level ELF codes
    # to that state's list of ELF Codes
    ex_us = elf_code_list[elf_code_list["Country Code (ISO 3166-1)"] != "US"]
    us_federal = elf_code_list[
        (elf_code_list["Country Code (ISO 3166-1)"] == "US")
        & (elf_code_list["Country sub-division code (ISO 3166-2)"].isnull())
    ]
    us_states = elf_code_list[
        (elf_code_list["Country Code (ISO 3166-1)"] == "US")
        & (elf_code_list["Country sub-division code (ISO 3166-2)"].notnull())
    ]

    def _assign_to_state(us_federal, state):
        us_federal_ = us_federal.copy()
        us_federal_["Country sub-division code (ISO 3166-2)"] = state
        return us_federal_

    # note: original federal level rows have been removed.
    return pandas.concat(
        [ex_us, us_states]
        + [
            _assign_to_state(us_federal, state)
            for state in us_states["Country sub-division code (ISO 3166-2)"].unique()
        ]
    )


class ELFAbbreviations:
    def __init__(self, elf_abbreviations_list):
        self._abbr_by_jurisdiction = elf_abbreviations_list.groupby("Jurisdiction")[
            "Abbreviation"
        ].apply(lambda abbreviations: list(numpy.unique(abbreviations)))
        self._elf_by_jur_and_abbr = elf_abbreviations_list.groupby(
            ["Jurisdiction", "Abbreviation"]
        )["ELF Code"].apply(lambda elf_codes: list(numpy.unique(elf_codes)))

    def abbreviations_for_jurisdiction(self, jurisdiction):
        return self._abbr_by_jurisdiction.get(jurisdiction, [])

    def elf_codes_for_abbreviation(self, jurisdiction, abbreviation):
        return self._elf_by_jur_and_abbr.get((jurisdiction, abbreviation), [])

    @staticmethod
    @lru_cache(maxsize=None)
    def matches(legal_name, abbr, use_lowercasing=True, use_endswith=True):
        if use_lowercasing:
            abbr = abbr.lower()
            legal_name = legal_name.lower()

        if use_endswith:
            if legal_name.endswith(" " + abbr):
                return True
        else:
            if abbr in legal_name:
                return True
        return False

    @staticmethod
    def from_elf_code_list(elf_code_list):
        elf_code_list = expand_us_states(elf_code_list)

        elf_abbreviations = elf_code_list.assign(
            Jurisdiction=get_jurisdiction(elf_code_list)
        )[["Jurisdiction", "ELF Code", "Abbreviations Local language"]].dropna()

        elf_abbreviations = (
            elf_abbreviations.assign(
                Abbreviation=lambda d: d["Abbreviations Local language"].str.split(";")
            )
            .explode("Abbreviation")
            .drop_duplicates(["Jurisdiction", "ELF Code", "Abbreviation"])
        )

        return ELFAbbreviations(elf_abbreviations)



import logging
from pathlib import Path

import joblib  # type: ignore
import pandas  # type: ignore
from sklearn.compose import ColumnTransformer  # type: ignore
from sklearn.feature_extraction.text import CountVectorizer  # type: ignore
from sklearn.metrics import accuracy_score, balanced_accuracy_score  # type: ignore
from sklearn.model_selection import train_test_split  # type: ignore
from sklearn.naive_bayes import ComplementNB  # type: ignore
from sklearn.pipeline import Pipeline  # type: ignore

from lenu.data import DataRepo
from lenu.data.lei import COL_LEGALNAME, COL_ELF
from lenu.ml.cnames import tokenize
from lenu.ml.features import ELFAbbreviationTransformer

logger = logging.getLogger(__name__)


def DefaultPipeline(elf_abbreviations):
    feature_extractor = ColumnTransformer(
        transformers=[
            (
                "abbreviations",
                ELFAbbreviationTransformer(elf_abbreviations=elf_abbreviations),
                [COL_LEGALNAME, COL_ELF, "Jurisdiction"],
            ),
            (
                "tokenizer",
                CountVectorizer(tokenizer=tokenize, lowercase=False, binary=True),
                COL_LEGALNAME,
            ),
        ]
    )
    pipeline_extPrep = Pipeline(
        steps=[
            ("feature_extraction", feature_extractor),
            ("classifier", ComplementNB()),
        ]
    )
    return pipeline_extPrep


def filter_infrequent_elf_codes(jurisdiction_data):
    # This fixes:
    # "ValueError: The least populated class in y has only 1 member, which is too few."
    filtered = jurisdiction_data.groupby(COL_ELF).filter(lambda x: len(x) >= 2)

    removed = jurisdiction_data[~jurisdiction_data[COL_ELF].isin(filtered[COL_ELF])]
    if len(removed) > 0:
        removed_elfs = list(removed[COL_ELF].unique())
        logger.warning(
            f"ELF Codes have been removed that appear only once: {removed_elfs}"
        )

    return filtered


def train_for_jurisdiction(jurisdiction_data, pipeline, test_size=1.0 / 3):
    filtered = filter_infrequent_elf_codes(jurisdiction_data)

    X = filtered[[COL_LEGALNAME, COL_ELF, "Jurisdiction"]]
    y = filtered[COL_ELF]

    # The minimum number of groups for any class cannot be less than 2.
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y)

    pipeline.fit(X_train, y_train)

    y_pred = pipeline.predict(X_test)

    accuracy = accuracy_score(y_true=y_test, y_pred=y_pred)
    logger.info(f"Model Accuracy: {accuracy}")
    bal_accuracy = balanced_accuracy_score(y_true=y_test, y_pred=y_pred)
    logger.info(f"Model Balanced Accuracy: {bal_accuracy}")

    return pipeline


class ELFDetectionModel:
    def __init__(self, jurisdiction, pipeline):
        self.jurisdiction = jurisdiction
        self.pipeline = pipeline

    def detect(self, legal_name, top=3):
        # preparing the input so that it fits
        input = pandas.DataFrame(
            {
                COL_LEGALNAME: [legal_name],
                COL_ELF: [""],
                "Jurisdiction": [self.jurisdiction],
            }
        )

        # do the prediction
        elf_probabilities = (
            pandas.Series(
                self.pipeline.predict_proba(input)[0], index=self.pipeline.classes_
            )
            .sort_values(ascending=False)
            .head(top)
        )

        return elf_probabilities


class ModelRepo:
    def __init__(self, models_dir: Path):
        self.models_dir = models_dir

    def train_pipeline(self, jurisdiction, data_loader: DataRepo):
        jurisdiction_data = data_loader.load_lei_cdf_data(jurisdiction)
        elf_abbreviations = data_loader.load_elf_abbreviations()

        pipeline = DefaultPipeline(elf_abbreviations)

        nsamples = len(jurisdiction_data)
        logger.info(
            f"Train model for jurisdiction {jurisdiction} ({nsamples} samples) ..."
        )
        pipeline = train_for_jurisdiction(jurisdiction_data, pipeline)

        model_file = self.models_dir.joinpath(f"complement_nb_{jurisdiction}.joblib")
        logger.info(f"Store model to {self.models_dir} ...")
        joblib.dump(pipeline, model_file)

    def get_model(self, jurisdiction) -> ELFDetectionModel:
        model_file = self.models_dir.joinpath(f"complement_nb_{jurisdiction}.joblib")

        if not model_file.exists():
            raise ValueError(
                f"No model for Jurisdiction {jurisdiction} in {self.models_dir}"
            )

        pipeline = joblib.load(model_file)

        return ELFDetectionModel(jurisdiction, pipeline)

    @staticmethod
    def from_models_dir(models_dir: Path) -> "ModelRepo":
        if not models_dir.exists() and not models_dir.is_dir():
            raise ValueError("Given model_dir does not exist or is not a directory")

        return ModelRepo(models_dir)
